
#include <SPI.h>
#include <SdFat.h>
#include <SamdAudio.h>
#include <Servo.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Adafruit_NeoPixel_ZeroDMA.h>
        
SdFat SD;
SamdAudio AudioPlayer;

Servo servo1, servo2;  

Adafruit_SSD1306 display(-1); // Initialize I2C display. -1 indicates no reset pin needed.

int ammo = 8; // Amount of ammo remaining.
int weapon = 0; // Currently selected weapon.
int dnacheck = 0; // DNA check state.  If 0 then the weapon can fire normally. If anything else, we're waiting for the DNA check to end.
unsigned long dnastart = 0; // Time at which we began the DNA check. Used to animate the progress bar and text.

unsigned long lastTime;
unsigned long time;
unsigned long timeDelta; 
float timeDeltaSec; 

#define PIN        11
#define NUM_PIXELS 60

Adafruit_NeoPixel_ZeroDMA strip(NUM_PIXELS, PIN, NEO_GRB); 


void I2C_write(byte addr, byte reg, byte val) { // Writes a byte to the specified register of the specified I2C device.
  Wire.beginTransmission(addr); 
  Wire.write(reg);
  Wire.write(val);               
  Wire.endTransmission();   
}


class LED {
  
  public:
  
    float current;  // Current LED brightness. Changes over time to match target.
    float target;   // Target LED brightness.  
    float scale;    // Scaling factor for LED brightness. Used when you want to be able to adjust the brightness of the LEDs without changing all the code that assumes the max desired brightness is 1.0.       
    float rise;     // How fast LED moves towards target when increasing. 10.0 = Starting from 0, LED would reach full brightness in 1/10th of a second. Set speed to -1 to change brightness instantly.                       
    float fall;     // How fast LED moves towards target when decreasing. Rise and fall times can be different so LEDs can fade in quickly, but fade out slowly to leave trails behind them.
 
    LED() {
      current = 0;
      target = 0;
      scale = 1;
      rise = -1; 
      fall = -1;
    }
    
    void set(float newTarget, boolean instant = false) {
      
      target = newTarget; 
      if (target < 0) { target = 0; } else if (target > 1) { target = 1; } // Clamp target to valid range.
      
      if ((instant) || ((current < target) && (rise < 0)) || ((current > target) && (fall < 0))) { // If instant is true, or the the speed of the LED in the specified direction is -1, change the LED's brightness level immediately.
        current = target; // Set LED immediately to desired brightness.
      }  
      
    }
    
    float get() { // Returns LED's current brightness, adjusted for scale and squared to make a linear ramp look correct to your eyes which perceive brightness in a logarithmic way.   
      return (current*current * scale*scale);
    }
    
    void toggle(float newTarget = 1.0, boolean instant = false) {
      if (target == 0) { set(newTarget, instant); } else { set(0, instant); } // Checks the LED's target value so even if the LED has not reached 0 yet, toggling will still work as expected.
    }
   
    void animate(float timePassed) { // Moves LED towards target brightness over time. timePassed = Seconds since last update.
      
      if (current < target) { // If brightness is increasing...
        if (rise > 0) { // If rise speed isn't 0 or -1 (instant)
          current += rise * timePassed; // Increase LED brightness according to how much time has passed between this update and the last.
          if (current > target) { current = target; } // Make sure we don't go past the target value.
        }
      } else if (current > target) { // If brightness is decreasing...
        if (fall > 0) { // If fall speed isn't 0 or -1 (instant)        
          current -= fall * timePassed; // Decrease LED brightness according to how much time has passed.
          if (current < target) { current = target; } // Make sure we don't go past the target value.
        }
      } 
      
    }  
         
};

class TLC59116 {
  
  public:
  
    byte address; 
    
    LED led[16]; 
     
    TLC59116(byte addr) { 
      address = addr;
    }
      
    void update(float timePassed) { // timePassed = Seconds since last update.
        
      // Enable oscillator and set outputs to allow PWM. (This really only needs to be done once when the module is initialized.)
        
        I2C_write(address, 0x00, 0x00); 
        
        I2C_write(address, 0x14, 0xFF); 
        I2C_write(address, 0x15, 0xFF);
        I2C_write(address, 0x16, 0xFF);
        I2C_write(address, 0x17, 0xFF);
        
      // Animate LEDs and Set duty cycle for outputs. Each output is updated immediately.
       
        for (int n = 0; n < 16; n++) {
          led[n].animate(timePassed);          
          I2C_write(address, 0x02 + n, 0xFF * led[n].get());
        }
                    
    }  
      
} LEDmodule1(0b1100000); 


void setup() {

  pinMode(0, INPUT_PULLDOWN); // Trigger
  pinMode(1, INPUT_PULLDOWN); // DNA check
  pinMode(2, INPUT_PULLDOWN); // Weapon select
  
  // Pins 0..5 are the LEDs. I have used 1K resistors so I can drive 2 LEDs from each pin.
  //for (int n=0; n<=5; n++) { pinMode(n, OUTPUT); }

  digitalWrite(26, LOW);
  pinMode(26, OUTPUT); // Writing to pin 26 turns on both the NFET and the Tau's red status LED. It can also be accessed with PIN_LED_TXL.  

  // Attach servos to pin 7 + 8.

    servo1.attach(7); 
    servo2.attach(8); 

    servo1.write(20); // Min & max angles will vary from servo to servo. For this servo valid angles are 20..160.  Outside this range the servo will continuously grind, use excessive power, and may eventually burn out.
    servo2.write(20); 

  if (!SD.begin(ATN)) { // Attempt to open SD card with CS on pin ATN.
    
    // Illuminate status LED momentarily (and trigger vibration motor and 3W LED) if we fail to read the SD card:
      digitalWrite(26, HIGH);
      delay(250);
      digitalWrite(26, LOW);
    
  }

  AudioPlayer.begin(22050);
  //AudioPlayer.play("fire.wav", 0); // Play startup sound.

  Wire.begin(); // Join I2C bus.
  
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // Initialize display with I2C address 0x3C. 
  display.clearDisplay(); // Clear the buffer. By default it has an Adafruit logo.
  display.display(); // Update the display. Not strictly necessary here.

  // Initialize Neopixel strip:
    strip.begin();
    strip.setBrightness(255);
    strip.show();

  // Initialize LEDs:
  
    LEDmodule1.led[4].rise = 4.0; // Set white LED so that it takes 1/4 second to fade in and out when it is turned on and off.
    LEDmodule1.led[4].fall = 4.0;

  time = millis(); 
  
}

void loop() {

  static int trigger = 0;
  static int select = 0;
          
  // Timing:
   
    lastTime = time;                           // Store start time of last update.
    time = millis();                           // Get current system time.
    timeDelta = time - lastTime;               // Calculate how long last update took, in milliseconds.
    timeDeltaSec = float(timeDelta) / 1000.0;  // Convert last update period into seconds.
      
  scrollPixels(3); // Scroll Neopixel strip left by 1 pixel, setting newly exposed pixel to black.
  
  if (dnacheck == 0) { // If we are not doing the DNA check right now...
  
    if (digitalRead(0) == HIGH) { // If trigger is being pulled...
  
      if (trigger == 0) { // And it wasn't being pulled the last time we checked...
  
        trigger = 1;
              
        if (ammo > 0) { 
          AudioPlayer.play("fire.wav", 0);
          digitalWrite(26, HIGH); // Trigger flash and vibration.
          servo2.write(160); // Move servo 2.
          ammo = ammo - 1;
        } else {
          AudioPlayer.play("empty.wav", 0);
        }
        
      }

      if (ammo > 0) { 
        strip.setPixelColor(0, colorWheel(51*weapon)); // Set first few pixels in Neopixel strip to color corresponding to currently selected weapon.
        strip.setPixelColor(1, colorWheel(51*weapon));
        strip.setPixelColor(2, colorWheel(51*weapon));
      }
      
    } else if (trigger == 1) { // If trigger is not being pulled, and it was...
      
      digitalWrite(26, LOW);
      servo2.write(20); // Reset servo 2 to default position.
      trigger = 0;
        
    }
    
    if (digitalRead(1) == HIGH) { // Select weapon: 
   
      if (select == 0) {
        
        select = 1;
        //AudioPlayer.play("select.wav", 0);
        
        weapon = weapon + 1;
        if (weapon > 4) { weapon = 0; }
  
        switch (weapon) {
          case 0: // Standard
            AudioPlayer.play("1.wav", 0);
            servo1.write(20); // Reset servo 1 to default position.
            break;
          case 1: // Heatseeker
            AudioPlayer.play("2.wav", 0);
            break;          
          case 2: // Armor piercing
            AudioPlayer.play("3.wav", 0);
            break;
          case 3: // Incendiary
            AudioPlayer.play("4.wav", 0);
            break;
          case 4: // High Ex
            AudioPlayer.play("5.wav", 0);
            servo1.write(160);
            break;
          /*
          case 5: // Ricochet
            AudioPlayer.play("6.wav", 0);
            servo1.write(20);
            break;
          */                                                       
        }
  
        ammo = 8;
  
      }
      
    } else if (select == 1) {
      select = 0;
    }

    if (digitalRead(2) == HIGH) { // Did user press the button to initiate a DNA check?  
      dnacheck = 1;
      dnastart = millis();
      AudioPlayer.play("dna.wav", 0);
    }
    
  } else { // DNA check in progress!

    // It takes 1330ms for the bar to fill. During this time it says DNA CHECK.  After that, it says I.D. FAIL until 6735ms have passed.
    // Animation is handled in the drawDisplay() function. 
    
  }

  // Update LED module LEDs:

    // Red and orange LEDs indicate currently selected weapon.
      if (weapon == 1) { LEDmodule1.led[0].set(1.0); } else { LEDmodule1.led[0].set(0); }
      if (weapon == 2) { LEDmodule1.led[1].set(1.0); } else { LEDmodule1.led[1].set(0); }
      if (weapon == 3) { LEDmodule1.led[2].set(1.0); } else { LEDmodule1.led[2].set(0); }
      if (weapon == 4) { LEDmodule1.led[3].set(1.0); } else { LEDmodule1.led[3].set(0); }

    // White LED indicates when out of ammo. This LED fades in and out over half a second.   
      if (ammo == 0) { LEDmodule1.led[4].set(1.0); } else { LEDmodule1.led[4].set(0); } 

    // RGB LED indicates color which will be used for Neopixels when you fire.
      LEDmodule1.led[5].set(float(colorWheelR(51*weapon))/255.0); 
      LEDmodule1.led[6].set(float(colorWheelG(51*weapon))/255.0);
      LEDmodule1.led[7].set(float(colorWheelB(51*weapon))/255.0);

    // Indicate ammo count on bargraph:
      
      for (int n = 15; n > 7; n--) {
        if (n >= (16-ammo)) { LEDmodule1.led[n].set(1.0); } else { LEDmodule1.led[n].set(0); }
      }

  strip.show(); // Update Neopixel strip using DMA.
  drawDisplay(); // Update OLED display via I2C.
  LEDmodule1.update(timeDeltaSec); // Animate LEDs and update LED module via I2C.
  
}


// This function scrolls the Neopixels by a specified number of pixels in either direction.
// Color data for the pixels scrolled past the end of the strip will be lost, while new pixels scrolled onto the strip will be off by default.
// You must call strip.show() after you call this function to display the changes. This enables you to fill the empty pixels with new color data before updating the strip, if desired.
void scrollPixels(int n) {

    if (n > 0) { // Scroll right.

      for(int i=strip.numPixels()-1; i >= 0; i--) { // Walk through strip backwards, replacing each pixel with the one n pixels before it. 
        if ((i-n) >= 0) { strip.setPixelColor(i, strip.getPixelColor(i-n)); } else { strip.setPixelColor(i, 0); }
      }

    } else { // Scroll left.

      for(int i=0; i < strip.numPixels(); i++) { // Walk through strip forwards, replacing each pixel with the one n pixels after it. 
        if ((i+n) < strip.numPixels()) { strip.setPixelColor(i, strip.getPixelColor(i+n)); } else { strip.setPixelColor(i, 0); }
      }
      
    }
   
}


// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
uint32_t colorWheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
    return strip.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
    return strip.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return strip.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
}

byte colorWheelR(byte WheelPos) {
  return (colorWheel(WheelPos) & 0xFF0000) >> 16; 
}

byte colorWheelG(byte WheelPos) {
  return (colorWheel(WheelPos) & 0xFF00) >> 8; 
}

byte colorWheelB(byte WheelPos) {
  return colorWheel(WheelPos) & 0xFF; 
}


void drawDisplay() {

  unsigned long timepassed;
  
  display.setTextSize(1);
  display.setTextColor(WHITE);
  
  display.clearDisplay(); // Clear the buffer. By default it has an Adafruit logo.

  // Line 0,20 to 127,20 above ammo counter.
    display.drawLine(0, 20, 127, 20, WHITE);
  
  // Lines between ammo counters at x = 24, 48, 72, 96 from y=21 to 31
    display.drawLine(24, 21, 24, 30, WHITE);
    display.drawLine(48, 21, 48, 30, WHITE);
    display.drawLine(72, 21, 72, 30, WHITE);
    display.drawLine(96, 21, 96, 30, WHITE);
    
  // Diagonal lines (Doesn't appear to be ammo display.)
  // 4 lines drawn bottom to top, 6x6 w/ 2px seperation, starting 24 pixels in from right, 2 pixels down from top.
    display.drawLine(107,    4, 113,   0, WHITE);
    display.drawLine(107+4,  4, 113+4, 0, WHITE);
    display.drawLine(107+8,  4, 113+8, 0, WHITE);
    display.drawLine(107+12, 4, 113+12, 0, WHITE);

  // Ammo type / DNA check:
    display.setCursor(0,9);

    if (dnacheck > 0) {

      // It takes 1330ms for the bar to fill. During this time it says DNA CHECK.  After that, it says I.D. FAIL until 6735ms have passed.

      timepassed = millis()-dnastart;
      
      if (timepassed > 6735) {
        dnacheck = 0;    
      } else if (timepassed > 1330) {
        display.println("I.D. FAIL");
        display.fillRect(0, 0, 101, 4, WHITE);
      } else {
        display.println("DNA CHECK");
        display.fillRect(0, 0, 101*(timepassed/1330.0), 4, WHITE);
        // Length of bar is maxlength*(timepassed/1330.0)
        // Bar stops just before SEMI and SEMI begins at x=103.
      }
        
    } else if (ammo == 0) {
      display.println("EMPTY");  
    } else {
      
      switch (weapon) {
        case 0: // Standard
          display.println("STANDARD");
          break;
        case 1: // Heatseeker
          display.println("HEATSEEKER");
          break;          
        case 2: // Armor piercing
          display.println("ARMOR PIERCING");
          break;
        case 3: // Incendiary
          display.println("INCENDIARY");
          break;
        case 4: // High Ex
          display.println("HIGH EX");
          break;
        case 5: // Ricochet
          display.println("RICOCHET");
          break;                                           
      }

    }
    
    // Semi-auto or rapid-fire:
      display.setCursor(103,9);
      display.println("SEMI");
 
  // Distance
    display.setCursor(0,22);
    display.println("D");
    display.setCursor(4,22);
    display.println(":");
    display.setCursor(6+3,22);
    display.println("0");
    display.setCursor(6+3+4,22);
    display.println(".");
    display.setCursor(6+3+6+3,22);
    display.println("0");
        
  // Ammunition
    if (ammo == 0) {
      display.setCursor(25,22);
      display.println("00ap");       
      display.setCursor(49,22);
      display.println("00in");
      display.setCursor(73,22);
      display.println("00he");    
      display.setCursor(98,22);
      display.println("00fmj");      
    } else {
      display.setCursor(25,22);
      display.println("25ap");       
      display.setCursor(49,22);
      display.println("25in");
      display.setCursor(73,22);
      display.println("25he");    
      display.setCursor(98,22);
      display.println("50fmj");
    }
    
  //display.setTextColor(BLACK, WHITE); // 'inverted' text
  //display.println(3.141592);
    
  display.display(); // Update the display. 

}


